from typing import Callable, List, Optional

import numpy as np

from ..circuit import Circuit
from ..gate import *
from .backendbase import Backend

_eye = np.eye(2, dtype=complex)


class OneQubitGateCompactionTranspiler(Backend):
    """Merge one qubit gate."""
    def _run_inner(self, gates, operations: List[Operation],
                   singlemats: List[np.ndarray], n_qubits: int,
                   mat1_decomposer: Callable[[OneQubitGate], List[Operation]]):
        for gate in gates:
            if isinstance(gate, Gate) and gate.n_qargs == 1:
                for t in gate.target_iter(n_qubits):
                    np.matmul(gate.matrix(), singlemats[t], out=singlemats[t])
            else:
                for t in gate.target_iter(n_qubits):
                    if not np.allclose(singlemats[t], _eye):
                        operations += mat1_decomposer(
                            Mat1Gate.create((t, ), (singlemats[t],)))
                        singlemats[t] = _eye.copy()
                operations.append(gate)

    def run(self,
            gates,
            n_qubits: int,
            *,
            mat1_decomposer: Optional[Callable[[OneQubitGate],
                                               List[Operation]]] = None,
            **kwargs) -> Circuit:
        """
        Args:
            basis (str or non-empty tuple of str): Name (lowername) of basis 2 qubits gate.
                If multiple basis gates are given, given basis gates are not decomposited and
                non-basis gate is decomposited to first available basis gate.
        """
        if mat1_decomposer is None:
            mat1_decomposer = lambda g: [g]
        operations = []
        singlemats = [_eye.copy() for _ in range(n_qubits)]
        self._run_inner(gates, operations, singlemats, n_qubits,
                        mat1_decomposer)
        for i, mat in enumerate(singlemats):
            if not np.allclose(mat, _eye):
                operations += mat1_decomposer(Mat1Gate.create((i, ), (mat,)))
        return Circuit(n_qubits, operations)



"""This module manages the set of operations, and provides a factory method of operations."""

from typing import Dict, Optional, Type

from . import gate
from .typing import Targets

GATE_SET: Dict[str, Type[gate.Operation]] = {
    # 1 qubit gates (alphabetical)
    "h": gate.HGate,
    "i": gate.IGate,
    "mat1": gate.Mat1Gate,
    "p": gate.PhaseGate,
    "phase": gate.PhaseGate,
    "r": gate.PhaseGate,
    "rx": gate.RXGate,
    "ry": gate.RYGate,
    "rz": gate.RZGate,
    "s": gate.SGate,
    "sdg": gate.SDagGate,
    "sx": gate.SXGate,
    "sxdg": gate.SXDagGate,
    "t": gate.TGate,
    "tdg": gate.TDagGate,
    "u": gate.UGate,
    "x": gate.XGate,
    "y": gate.YGate,
    "z": gate.ZGate,
    # Controlled gates (alphabetical)
    "ccx": gate.ToffoliGate,
    "ccz": gate.CCZGate,
    "cnot": gate.CXGate,
    "ch": gate.CHGate,
    "cp": gate.CPhaseGate,
    "cphase": gate.CPhaseGate,
    "cr": gate.CPhaseGate,
    "crx": gate.CRXGate,
    "cry": gate.CRYGate,
    "crz": gate.CRZGate,
    "cswap": gate.CSwapGate,
    "cu": gate.CUGate,
    "cx": gate.CXGate,
    "cy": gate.CYGate,
    "cz": gate.CZGate,
    "toffoli": gate.ToffoliGate,
    # Other multi qubit gates (alphabetical)
    "rxx": gate.RXXGate,
    "ryy": gate.RYYGate,
    "rzz": gate.RZZGate,
    "swap": gate.SwapGate,
    "zz": gate.ZZGate,
    # Measure and reset (alphabetical)
    "m": gate.Measurement,
    "measure": gate.Measurement,
    "reset": gate.Reset,
}

def get_op_type(name: str) -> Optional[Type[gate.Operation]]:
    """Get a class of operation from operation name."""
    return GATE_SET.get(name)

def create(name: str, targets: Targets, params: tuple) -> gate.Operation:
    """Create an operation from name, targets and params."""
    op_type = get_op_type(name)
    if op_type is None:
        raise ValueError(f"Unknown operation `{name}'.")
    return op_type.create(targets, params)

def register_operation(name: str, op_type: Type[gate.Operation]) -> None:
    """Register an operation. If operation is already exists, overwrite it."""
    GATE_SET[name] = op_type

def unregister_operation(name: str) -> None:
    """Unregister an operation. If operation is not exists, do nothing."""
    try:
        del GATE_SET[name]
    except KeyError:
        pass



from collections import Counter
import math
from typing import Any, List, Optional

import numpy as np

from .backendbase import Backend
import blueqat.gate as bqgate

class _U3(bqgate.OneQubitGate):
    lowername = '_u3'
    def __init__(self, targets, theta, phi, lam):
        super().__init__(targets, (theta, phi, lam))

class _CU3(bqgate.TwoQubitGate):
    lowername = '_cu3'
    def __init__(self, targets, theta, phi, lam):
        super().__init__(targets, (theta, phi, lam))

class _Expii(bqgate.OneQubitGate):
    lowername = '_expii'
    def __init__(self, targets, theta):
        super().__init__(targets, (theta,))

class _CExpii(bqgate.TwoQubitGate):
    lowername = '_cexpii'
    def __init__(self, targets, theta):
        super().__init__(targets, (theta,))

class QgateBackend(Backend):
    def __init__(self):
        import sys
        this = sys.modules[__name__]
        if not hasattr(this, 'qgate'):
            import qgate
            this.qgate = qgate
            this.model = qgate.model
            this.gtype = qgate.model.gate_type

            QgateBackend.gatetypes = {
                'i': (gtype.ID, 0, 1),
                'x': (gtype.X, 0, 1),
                'y': (gtype.Y, 0, 1),
                'z': (gtype.Z, 0, 1),
                'h': (gtype.H, 0, 1),
                't': (gtype.T, 0, 1),
                's': (gtype.S, 0, 1),
                # adjoint
                'tdg': (gtype.T, 0, 1),
                'sdg': (gtype.S, 0, 1),
                # 1-control-bit gate
                'cz': (gtype.Z, 1, 1),
                'cx': (gtype.X, 1, 1),
                'cnot': (gtype.X, 1, 1),
                # rotation, phase gate
                'rx': (gtype.RX, 0, 1),
                'ry': (gtype.RY, 0, 1),
                'rz': (gtype.RZ, 0, 1),
                'phase': (gtype.U1, 0, 1),
                # controlled rotation gate
                'crx': (gtype.RX, 1, 1),
                'cry': (gtype.RY, 1, 1),
                'crz': (gtype.RZ, 1, 1),
                'cphase': (gtype.U1, 1, 1),
                # U gate
                'u': None,
                '_u3': (gtype.U, 0, 1),
                # controlled U gate
                'cu': None,
                '_cu3': (gtype.U, 1, 1),
                # global phase
                '_expii': (gtype.ExpiI, 0, 1),
                '_cexpii': (gtype.ExpiI, 1, 1),
                # swap and multi-controlled-bit gate
                'swap': (gtype.SWAP, 0, 2),
                'ccx': (gtype.X, 2, 1),
                'ccz': (gtype.Z, 2, 1)
            }

            if hasattr(
                    qgate,
                    '__version__') and qgate.__version__.startswith('0.4.'):
                this.version = qgate.__version__
                # register new gate types in Qgate 0.4
                QgateBackend.gatetypes['rxx'] = (gtype.RXX, 0, 2)
                QgateBackend.gatetypes['ryy'] = (gtype.RYY, 0, 2)
                QgateBackend.gatetypes['rzz'] = (gtype.RZZ, 0, 2)
                QgateBackend.gatetypes['cphase'] = (gtype.P, 1, 1)
                QgateBackend.gatetypes['cswap'] = (gtype.SWAP, 1, 2)

                # wrappers to absorb interface changes for Qgate 0.4
                def set_qregs_04(gate, qreg, controls):
                    gate.set_target(qreg)
                    if controls:
                        gate.set_controls(controls)

                def get_target_04(gate):
                    return gate.target

                def create_multi_qubit_gate_04(gate, typeinfo, ctrls, targets):
                    g = model.Gate(typeinfo[0](*gate.params))
                    if ctrls:
                        g.set_controls(ctrls)
                    g.set_targets(targets)
                    return g

                def create_simulator_04(prefs):
                    return qgate.simulator.create(**prefs)

                def create_sampler_04(sim, qreg_ordering, sampling_method):
                    if sampling_method == 'blueqat':
                        # create prob numpy array
                        probs = sim.qubits.create_prob_accessor(
                            qreg_ordering)[:]
                        return BlueqatCompatibleSampler_04(
                            probs, qreg_ordering)
                    else:
                        return SamplingWrapper_04(sim.qubits, qreg_ordering)

                QgateBackend.set_qregs = set_qregs_04
                QgateBackend.get_target = get_target_04
                QgateBackend.create_multi_qubit_gate = create_multi_qubit_gate_04
                QgateBackend.create_simulator = create_simulator_04
                QgateBackend.create_sampler = create_sampler_04
            else:
                this.version = ''
                # wrappers to absorb interface changes for Qgate 0.3 or older
                def set_qregs_03(gate, qreg, controls):
                    gate.set_qreg(qreg)
                    if controls:
                        gate.set_ctrllist(controls)

                def get_target_03(gate):
                    return gate.qreg

                def create_multi_qubit_gate_03(gate, typeinfo, ctrls, targets):
                    assert not ctrls, 'controlled multi qubit gate is not supported'
                    g = model.MultiQubitGate(typeinfo[0](*gate.params))
                    g.set_qreglist(targets)
                    return g

                def create_simulator_03(prefs):
                    circuit_prep = prefs.get('circuit_prep',
                                             qgate.prefs.dynamic)
                    simprefs = {'circuit_prep': circuit_prep}
                    # create simulator instance
                    runtime = prefs.get('runtime', 'cpu')
                    if runtime == 'cuda':
                        sim = qgate.simulator.cuda(**simprefs)
                    elif runtime == 'cpu':
                        sim = qgate.simulator.cpu(**simprefs)
                    elif runtime == 'py':
                        sim = qgate.simulator.py(**simprefs)
                    else:
                        raise RuntimeError(
                            "Unknown runtime, {}.  Accetable values are 'cpu', 'cuda' and 'py'."
                            .format(runtime))
                    return sim

                def create_sampler_03(sim, qreg_ordering, sampling_method):
                    if sampling_method == 'blueqat':
                        # create prob numpy array
                        return sim.qubits.create_sampling_pool(
                            qreg_ordering, BlueqatCompatibleSampler_03)
                    else:
                        return sim.qubits.create_sampling_pool(qreg_ordering)

                QgateBackend.set_qregs = set_qregs_03
                QgateBackend.get_target = get_target_03
                QgateBackend.create_multi_qubit_gate = create_multi_qubit_gate_03
                QgateBackend.create_simulator = create_simulator_03
                QgateBackend.create_sampler = create_sampler_03

    def run(self,
            gates: List[bqgate.Operation],
            n_qubits: int,
            returns: Optional[str] = None,
            shots: Optional[int] = None,
            sampling: str = 'qgate',
            initial: Optional[np.ndarray] = None,
            **kwargs) -> Any:
        self.n_qubits = n_qubits
        r = returns or ''
        shots = shots or 0
        if shots != 0 and r == '':
            r = 'shots'
        elif r == '':
            r = 'statevector'

        sim = QgateBackend.create_simulator(kwargs)

        # creating registers and references.
        self.qregs = [model.Qreg() for _ in range(n_qubits)]
        self.refs = [model.Reference() for _ in range(n_qubits)]
        sim.qubits.set_ordering(self.qregs)
        if initial is not None:
            sim.qubits.add_qreg_group(self.qregs, initial)

        ops = self.convert_operators(gates)
        # qgate.dump(ops)
        if r == 'statevector':
            return self.get_state_vector(sim, ops)
        elif r == 'shots':
            return Counter(self.sample(sim, ops, shots, sampling))
        elif r == 'statevector_and_shots':
            if 1 < shots:
                raise RuntimeError(
                    'shots must be 1 when returns="statevector_and_shots"')
            vec, shots = self.get_state_vector_and_sample(sim, ops)
            return vec, Counter(shots)
        else:
            raise RuntimeError('Unknown returns token, {}.'.format(r))

    def convert_operators(self, gates):
        ops = list()
        for op in gates:
            typeinfo = QgateBackend.gatetypes.get(op.lowername, None)
            if typeinfo is not None:
                if typeinfo[2] == 1:
                    ops += self.convert_one_qubit_gate(op, typeinfo)
                else:
                    ops += self.convert_multi_qubit_gate(op, typeinfo)
            elif op.lowername == 'u':
                ops += self.convert_operators([
                    _U3(op.targets, op.theta, op.phi, op.lam),
                    _Expii(op.targets, op.gamma + 0.5 * (op.phi + op.lam))
                ])
            elif op.lowername == 'cu':
                ops += self.convert_operators([
                    _CU3(op.targets, op.theta, op.phi, op.lam),
                    _CExpii(op.targets, op.gamma + 0.5 * (op.phi + op.lam))
                ])
            elif op.lowername in ('measure', 'm'):
                # "measure": gate.Measurement, "m": gate.Measurement,
                ops += self.convert_measure(op)
            else:
                assert False
        return ops

    def convert_one_qubit_gate(self, gate, typeinfo):
        glist = list()

        if typeinfo[1] == 0:
            # no control bit
            assert isinstance(gate, bqgate.OneQubitGate)
            one_qubit_gates = self.create_gates(typeinfo[0], gate)
            # "tdg": gate.TDagGate, "sdg": gate.SDagGate,
            if gate.lowername[1:] == 'dg':
                for qg in one_qubit_gates:
                    qg.set_adjoint(True)
            glist += one_qubit_gates
        else:
            if isinstance(gate, bqgate.TwoQubitGate):
                # controlled gate
                for ctrlidx, qregidx in gate.control_target_iter(
                        self.n_qubits):
                    gt = typeinfo[0](*gate.params)
                    g = model.Gate(gt)
                    ctrl = self.qregs[ctrlidx]
                    qreg = self.qregs[qregidx]
                    QgateBackend.set_qregs(g, qreg, [ctrl])
                    glist.append(g)
            elif isinstance(gate, bqgate.ToffoliGate):
                gt = typeinfo[0](*gate.params)
                g = model.Gate(gt)
                c0, c1, t = gate.targets
                ctrls = (self.qregs[c0], self.qregs[c1])
                qreg = self.qregs[t]
                QgateBackend.set_qregs(g, qreg, ctrls)
                glist.append(g)
            else:
                raise RuntimeError('Unknown gate, {}.'.format(repr(gate)))

        return glist

    def convert_multi_qubit_gate(self, gate, typeinfo):
        glist = list()
        n_ctrls = typeinfo[1]
        ctrls = gate.targets[:n_ctrls]
        ctrl_qregs = [self.qregs[idx] for idx in ctrls]
        targets = gate.targets[n_ctrls:]
        target_qregs = [self.qregs[idx] for idx in targets]
        g = QgateBackend.create_multi_qubit_gate(gate, typeinfo, ctrl_qregs,
                                                 target_qregs)
        glist.append(g)
        return glist

    def convert_measure(self, measure):
        mlist = list()
        for qregidx in measure.target_iter(self.n_qubits):
            m = model.Measure(self.refs[qregidx], self.qregs[qregidx])
            mlist.append(m)
        return mlist

    def create_gates(self, gate_type_factory, gate):
        glist = list()
        for qregidx in gate.target_iter(self.n_qubits):
            gt = gate_type_factory(*gate.params)
            g = model.Gate(gt)
            qreg = self.qregs[qregidx]
            QgateBackend.set_qregs(g, qreg, None)
            glist.append(g)
        return glist

    def create_global_phase_gates(self, phase, gate):
        glist = list()
        for qregidx in gate.target_iter(self.n_qubits):
            gt = gtype.ExpiI(phase)
            g = model.Gate(gt)
            qreg = self.qregs[qregidx]
            self.set_qregs(g, qreg, None)
            glist.append(g)
        return glist

    def get_state_vector(self, sim, ops):
        sim.run(ops)
        return sim.qubits.states[:]

    def sample(self, sim, ops, shots, sampling_method):
        sampling_qregs = list()
        ops_no_measure = list()

        for op in ops:
            if isinstance(op, model.Measure):
                target = QgateBackend.get_target(op)
                sampling_qregs.append(target)
            else:
                ops_no_measure.append(op)

        sim.run(ops_no_measure)
        # print(sim.qubits.states[:])

        qregs_no_dup = set(sampling_qregs)
        sampling_qregs = list(qregs_no_dup)
        sampling_qregs.sort(key=self.qregs.index)

        sampler = QgateBackend.create_sampler(sim, sampling_qregs,
                                              sampling_method)
        obs = sampler.sample(shots)

        # FIXME: faster conversion for bit representation
        hist = obs.histgram()
        bit_format = '0{}b'.format(len(self.qregs))
        sampling_pos = [self.qregs.index(qreg) for qreg in sampling_qregs]
        max_bit_pos = len(self.qregs) - 1
        strkey_hist = dict()
        for v, c in hist.items():
            permuted = 0
            for idx, pos in enumerate(sampling_pos):
                if (v & (1 << idx)) != 0:
                    permuted |= 1 << (max_bit_pos - pos)
            strbits = format(permuted, bit_format)
            strkey_hist[strbits] = c
        return strkey_hist

    def get_state_vector_and_sample(self, sim, ops):
        sim.run(ops)
        # state vector
        state_vector = sim.qubits.states[:]
        # shot
        reversed_refs = list(reversed(self.refs))
        obs = sim.obs(reversed_refs)
        bit_format = '0{}b'.format(len(self.refs))
        strkey_hist = dict()
        strbits = format(obs.int, bit_format)
        strkey_hist[strbits] = 1  # observed 1 time.
        return state_vector, strkey_hist


# provided for compatiblity and tests
class BlueqatCompatibleSampler_03:
    def __init__(self, prob, empty_lanes, qreg_ordering):
        #prob *= (1. / np.sum(prob))
        self.prob = prob

        self.qreg_ordering = qreg_ordering
        self.empty_lanes = empty_lanes
        self.mask = 0
        for idx in empty_lanes:
            self.mask |= 1 << idx

    def sample(self, n_samples):
        obs = np.empty([n_samples], np.int)
        import random
        for idx in range(n_samples):
            redprob = self.prob
            v = 0
            p_all = 1
            for bitpos in range(len(self.qreg_ordering)):
                rnum = random.random()
                p_zero = np.sum(redprob[::2])
                if rnum < (p_zero / p_all):
                    redprob = redprob[::2]
                    p_all = p_zero
                else:
                    redprob = redprob[1::2]
                    p_all = p_all - p_zero
                    v |= (1 << bitpos)

            obs[idx] = v

        if self.mask != 0:
            self.shift_for_empty_lanes(obs)
        obslist = qgate.simulator.observation.ObservationList(
            self.qreg_ordering, obs, self.mask)
        return obslist

    def shift_for_empty_lanes(self, obs):
        # create control bit mask
        bits = [1 << lane for lane in self.empty_lanes]
        n_shifts = len(bits) + 1

        mask = bits[0] - 1
        masks = [mask]
        for idx in range(len(bits) - 1):
            mask = (bits[idx + 1] - 1) & (~(bits[idx] * 2 - 1))
            masks.append(mask)
        mask = ~(bits[-1] * 2 - 1)
        masks.append(mask)

        for idx in range(len(obs)):
            v = obs[idx]
            v_shifted = 0
            for shift in range(n_shifts):
                v_shifted |= (v << shift) & masks[shift]
            obs[idx] = v_shifted


# provided for compatiblity and tests
class BlueqatCompatibleSampler_04:
    def __init__(self, prob, qreg_ordering):
        self.prob = prob
        self.qreg_ordering = qreg_ordering

    def sample(self, n_samples):
        obs = np.empty([n_samples], np.int)
        import random
        for idx in range(n_samples):
            redprob = self.prob
            v = 0
            p_all = 1
            for bitpos in range(len(self.qreg_ordering)):
                rnum = random.random()
                p_zero = np.sum(redprob[::2])
                if rnum < (p_zero / p_all):
                    redprob = redprob[::2]
                    p_all = p_zero
                else:
                    redprob = redprob[1::2]
                    p_all = p_all - p_zero
                    v |= (1 << bitpos)

            obs[idx] = v

        obslist = qgate.simulator.observation.ObservationList(
            self.qreg_ordering, obs, 0)
        return obslist


class SamplingWrapper_04:
    def __init__(self, qubits, qreg_ordering):
        self.qubits = qubits
        self.qreg_ordering = qreg_ordering

    def sample(self, n_samples, randnums=None):
        return self.qubits.sample(self.qreg_ordering, n_samples, randnums)



# Copyright 2019 The Blueqat Developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import warnings
from functools import reduce
from .backendbase import Backend


def lazy_import():
    global eye, symbols, sin, cos, exp, sqrt, pi, I, Matrix, sympy_gate, TensorProduct, sympy
    from sympy import eye, symbols, sin, cos, exp, sqrt, pi, I, Matrix
    from sympy.physics.quantum import TensorProduct
    import sympy


def _angle_simplify(ang):
    if isinstance(ang, float):
        nsimp = sympy.nsimplify(ang / np.pi)
        numer, denom = nsimp.as_numer_denom()
        if denom < 1e12:
            return pi * numer / denom
    return ang


class _SympyBackendContext:
    def __init__(self, n_qubits, ignore_global):
        self.n_qubits = n_qubits
        self.matrix_of_circuit = eye(2**n_qubits)
        self.ignore_global = ignore_global


class SympyBackend(Backend):
    def __init__(self):
        try:
            lazy_import()
        except ImportError:
            raise ImportError(
                'sympy_unitary requires sympy. Please install before call this option.'
            ) from None
        theta, phi, lam, gamma = symbols('theta phi lam gamma')
        self.theta = theta
        self.phi = phi
        self.lam = lam
        self.gamma = gamma
        self.SYMPY_GATE = {
            '_C0':
            Matrix([[1, 0], [0, 0]]),
            '_C1':
            Matrix([[0, 0], [0, 1]]),
            'H':
            Matrix([[1, 1], [1, -1]]) / sqrt(2),
            'PHASE':
            Matrix([[1, 0], [0, exp(I * theta)]]),
            'RX':
            Matrix([[cos(theta / 2), -I * sin(theta / 2)],
                    [-I * sin(theta / 2), cos(theta / 2)]]),
            'RY':
            Matrix([[cos(theta / 2), -sin(theta / 2)],
                    [sin(theta / 2), cos(theta / 2)]]),
            'RZ':
            Matrix([[cos(theta / 2) - I * sin(theta / 2), 0],
                    [0, cos(theta / 2) + I * sin(theta / 2)]]),
            'SX':
            Matrix([[1 + I, 1 - I], [1 - I, 1 + I]]) / 2,
            'SXDG':
            Matrix([[1 - I, 1 + I], [1 + I, 1 - I]]) / 2,
            'U':
            Matrix(
                [[cos(theta / 2), -(cos(lam) + I * sin(lam)) * sin(theta / 2)],
                 [(cos(phi) + I * sin(phi)) * sin(theta / 2),
                  (cos(phi + lam) + I * sin(phi + lam)) * cos(theta / 2)]]) *
            (cos(gamma) + I * sin(gamma)),
            'X':
            Matrix([[0, 1], [1, 0]]),
            'Y':
            Matrix([[0, -I], [I, 0]]),
            'Z':
            Matrix([[1, 0], [0, -1]]),
        }

    def _create_matrix_of_one_qubit_gate(self, n_qubits, targets,
                                         matrix_of_gate):
        targets = [idx for idx in targets]
        gates = []
        for idx in range(n_qubits):
            if idx in targets:
                gates.append(matrix_of_gate)
            else:
                gates.append(eye(2))
        return reduce(TensorProduct, reversed(gates))

    def _create_matrix_of_one_qubit_gate_circuit(self, gate, ctx,
                                                 matrix_of_gate):
        n_qubits = ctx.n_qubits
        m = self._create_matrix_of_one_qubit_gate(n_qubits,
                                                  gate.target_iter(n_qubits),
                                                  matrix_of_gate)
        ctx.matrix_of_circuit = m * ctx.matrix_of_circuit
        return ctx

    def _one_qubit_gate_noargs(self, gate, ctx):
        matrix_of_gate = self.SYMPY_GATE[gate.uppername]
        return self._create_matrix_of_one_qubit_gate_circuit(
            gate, ctx, matrix_of_gate)

    def _one_qubit_gate_args_theta(self, gate, ctx):
        theta = _angle_simplify(gate.theta)
        matrix_of_gate = self.SYMPY_GATE[gate.uppername].subs(
            self.theta, theta)
        return self._create_matrix_of_one_qubit_gate_circuit(
            gate, ctx, matrix_of_gate)

    gate_x = _one_qubit_gate_noargs
    gate_y = _one_qubit_gate_noargs
    gate_z = _one_qubit_gate_noargs
    gate_h = _one_qubit_gate_noargs
    gate_sx = _one_qubit_gate_noargs
    gate_sxdg = _one_qubit_gate_noargs
    gate_rx = _one_qubit_gate_args_theta
    gate_ry = _one_qubit_gate_args_theta
    gate_rz = _one_qubit_gate_args_theta
    gate_phase = _one_qubit_gate_args_theta

    def gate_u(self, gate, ctx):
        theta, phi, lam, gamma = map(_angle_simplify, gate.params)
        matrix_of_gate = self.SYMPY_GATE['U'].subs(
            [(self.lam, lam), (self.phi, phi), (self.theta, theta),
             (self.gamma, gamma)],
            simultaneous=True)
        return self._create_matrix_of_one_qubit_gate_circuit(
            gate, ctx, matrix_of_gate)

    def _create_matrix_of_two_qubit_gate(self,
                                         n_qubits,
                                         gate,
                                         control,
                                         target,
                                         subs=None):
        matrix_of_target_gate = self.SYMPY_GATE[gate.uppername[1:]]
        if subs:
            matrix_of_target_gate = matrix_of_target_gate.subs(
                subs, simultaneous=True)
        c0 = self._create_matrix_of_one_qubit_gate(n_qubits, [control],
                                                   self.SYMPY_GATE['_C0'])
        c1 = self._create_matrix_of_one_qubit_gate(n_qubits, [control],
                                                   self.SYMPY_GATE['_C1'])
        tgt = self._create_matrix_of_one_qubit_gate(n_qubits, [target],
                                                    matrix_of_target_gate)
        return c0 + tgt * c1

    def _two_qubit_gate_noargs(self, gate, ctx):
        n_qubits = ctx.n_qubits
        for control, target in gate.control_target_iter(n_qubits):
            m = self._create_matrix_of_two_qubit_gate(n_qubits, gate, control,
                                                      target)
            ctx.matrix_of_circuit = m * ctx.matrix_of_circuit
        return ctx

    gate_cx = _two_qubit_gate_noargs
    gate_cz = _two_qubit_gate_noargs

    def gate_cu(self, gate, ctx):
        theta, phi, lam, gamma = map(_angle_simplify, gate.params)
        n_qubits = ctx.n_qubits
        subs = [(self.lam, lam), (self.phi, phi), (self.theta, theta),
                (self.gamma, gamma)]
        for control, target in gate.control_target_iter(n_qubits):
            m = self._create_matrix_of_two_qubit_gate(n_qubits, gate, control,
                                                      target, subs)
            ctx.matrix_of_circuit = m * ctx.matrix_of_circuit
        return ctx

    def _two_qubit_gate_args_theta(self, gate, ctx):
        theta = _angle_simplify(gate.theta)
        matrix_of_gate = self.SYMPY_GATE[gate.uppername].subs(
            self.theta, theta)
        for control, target in gate.control_target_iter(ctx.n_qubits):
            control_gate_of_matrix = self._create_matrix_of_two_qubit_gate(
                gate, ctx, control, target)
            ctx.matrix_of_circuit = control_gate_of_matrix * ctx.matrix_of_circuit
        return ctx

    def _warn_unavailable_operation(self, gate, ctx):
        warnings.warn(
            f'`{gate.uppername}` operation is ignored because unavailable in this backend.'
        )
        return ctx

    gate_measure = _warn_unavailable_operation
    gate_reset = _warn_unavailable_operation

    def _preprocess_run(self, gates, n_qubits, args, kwargs):
        kwargs.setdefault('ignore_global', False)
        return gates, _SympyBackendContext(n_qubits, kwargs['ignore_global'])

    def _postprocess_run(self, ctx):
        if ctx.ignore_global:
            mat = ctx.matrix_of_circuit
            for i in range(2**ctx.n_qubits):
                if mat[i, i] != 0:
                    return mat * (abs(mat[i, i]) / mat[i, i])
        return ctx.matrix_of_circuit



# Copyright 2019 The Blueqat Developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This module defines Circuit and the setting for circuit.
"""

import warnings
from functools import partial
import typing
from typing import Callable, Dict, Optional, Tuple, Type

import numpy as np

from . import gate
from .gateset import get_op_type, register_operation, unregister_operation

if typing.TYPE_CHECKING:
    from .gate import Operation
    from .backends.backendbase import Backend
    BackendUnion = typing.Union[None, str, Backend]

GLOBAL_MACROS = {}


class Circuit:
    """Store the gate operations and call the backends."""
    def __init__(self, n_qubits=0, ops=None):
        self.ops = ops or []
        self._backends: Dict[str, 'Backend'] = {}
        self.n_qubits = n_qubits

    def __repr__(self):
        return f'Circuit({self.n_qubits}).' + '.'.join(
            str(op) for op in self.ops)

    def __get_backend(self, backend_name: str) -> 'Backend':
        try:
            return self._backends[backend_name]
        except KeyError:
            backend = BACKENDS.get(backend_name)
            if backend is None:
                raise ValueError(f"Backend {backend_name} doesn't exist.")
        self._backends[backend_name] = backend()
        return self._backends[backend_name]

    def __backend_runner_wrapper(self, backend_name):
        backend = self.__get_backend(backend_name)

        def runner(*args, **kwargs):
            return backend.run(self.ops, self.n_qubits, *args, **kwargs)

        return runner

    def __getattr__(self, name):
        op_type = get_op_type(name)
        if op_type:
            return _GateWrapper(self, op_type)
        if name in GLOBAL_MACROS:
            return partial(GLOBAL_MACROS[name], self)
        if name.startswith("run_with_"):
            backend_name = name[9:]
            if backend_name in BACKENDS:
                return self.__backend_runner_wrapper(backend_name)
            raise AttributeError(f"Backend '{backend_name}' is not exists.")
        raise AttributeError(
            f"'circuit' object has no attribute or gate '{name}'")

    def __add__(self, other):
        if not isinstance(other, Circuit):
            return NotImplemented
        c = self.copy()
        c += other
        return c

    def __iadd__(self, other):
        if not isinstance(other, Circuit):
            return NotImplemented
        self.ops += other.ops
        self.n_qubits = max(self.n_qubits, other.n_qubits)
        return self

    def copy(self,
             copy_backends: bool = True) -> 'Circuit':
        """Copy the circuit.

        params:
            | copy_backends :bool copy backends if True.
        """
        copied = Circuit(self.n_qubits, self.ops.copy())
        if copy_backends:
            copied._backends = {k: v.copy() for k, v in self._backends.items()}
        return copied

    def dagger(self,
               ignore_measurement: bool = False) -> 'Circuit':
        """Make Hermitian conjugate of the circuit.

        This feature is beta. Interface may be changed.

        ignore_measurement (bool, optional):
            | If True, ignore the measurement in the circuit.
            | Otherwise, if measurement in the circuit, raises ValueError.
        """
        ops = []
        for g in reversed(self.ops):
            try:
                ops.append(g.dagger())
            except ValueError:
                if not ignore_measurement:
                    raise ValueError(
                        'Cannot make the Hermitian conjugate of this circuit because '
                        'the circuit contains measurement.')

        copied = Circuit(self.n_qubits, ops)
        return copied

    def run(self, *args, backend=None, **kwargs):
        """Run the circuit.

        `Circuit` have several backends. When `backend` parameter is specified,
        use specified backend, and otherwise, default backend is used.
        Other parameters are passed to the backend.

        The meaning of parameters are depends on the backend specifications.
        However, following parameters are commonly used.

        Commonly used args (Depends on backend):
            | shots (int, optional): The number of sampling the circuit.
            | returns (str, optional):  The category of returns value.
            |   e.g. "statevector" returns the state vector after run the circuit.
            |         "shots" returns the counter of measured value.
            | token, url (str, optional): The token and URL for cloud resource.

        Returns:
            Depends on backend.

        Raises:
            Depends on backend.
        """
        if backend is None:
            backend = self.__get_backend(DEFAULT_BACKEND_NAME)
        elif isinstance(backend, str):
            backend = self.__get_backend(backend)
        return backend.run(self.ops, self.n_qubits, *args, **kwargs)

    def make_cache(self, backend: Optional['BackendUnion'] = None) -> None:
        """Make a cache to reduce the time of run. Some backends may implemented it.

        This is temporary API. It may changed or deprecated."""
        if backend is None:
            backend = DEFAULT_BACKEND_NAME
        if isinstance(backend, str):
            backend = self.__get_backend(backend)
        return backend.make_cache(self.ops, self.n_qubits)

    def to_qasm(self, *args, **kwargs):
        """Returns the OpenQASM output of this circuit."""
        return self.run_with_qasm_output(*args, **kwargs)

    def to_unitary(self, *args, **kwargs):
        """Returns sympy unitary matrix of this circuit."""
        return self.run_with_sympy_unitary(*args, **kwargs)

    def statevector(self,
                    backend: 'BackendUnion' = None,
                    **kwargs) -> np.ndarray:
        """Run the circuit and get a statevector as a result."""
        if kwargs.get('returns'):
            raise ValueError('Circuit.statevector has no argument `returns`.')
        if backend is None:
            backend = self.__get_backend(DEFAULT_BACKEND_NAME)
        elif isinstance(backend, str):
            backend = self.__get_backend(backend)

        if hasattr(backend, 'statevector'):
            return backend.statevector(self.ops, self.n_qubits,
                                       **kwargs)
        return backend.run(self.ops,
                           self.n_qubits,
                           returns='statevector',
                           **kwargs)

    def shots(self,
              shots: int,
              backend: 'BackendUnion' = None,
              **kwargs) -> typing.Counter[str]:
        """Run the circuit and get shots as a result."""
        if kwargs.get('returns'):
            raise ValueError('Circuit.shots has no argument `returns`.')
        if backend is None:
            backend = self.__get_backend(DEFAULT_BACKEND_NAME)
        elif isinstance(backend, str):
            backend = self.__get_backend(backend)

        if hasattr(backend, 'shots'):
            return backend.shots(self.ops,
                                 self.n_qubits,
                                 shots=shots,
                                 **kwargs)
        return backend.run(self.ops,
                           self.n_qubits,
                           shots=shots,
                           returns='shots',
                           **kwargs)

    def oneshot(self,
                backend: 'BackendUnion' = None,
                **kwargs) -> Tuple[np.ndarray, str]:
        """Run the circuit and get shots as a result."""
        if kwargs.get('returns'):
            raise ValueError('Circuit.oneshot has no argument `returns`.')
        if backend is None:
            backend = self.__get_backend(DEFAULT_BACKEND_NAME)
        elif isinstance(backend, str):
            backend = self.__get_backend(backend)

        if hasattr(backend, 'oneshot'):
            return backend.oneshot(self.ops, self.n_qubits, **kwargs)
        v, cnt = backend.run(self.ops,
                             self.n_qubits,
                             shots=1,
                             returns='statevector_and_shots',
                             **kwargs)
        return v, cnt.most_common()[0][0]


class _GateWrapper:
    def __init__(self, circuit: Circuit, op_type: Type['Operation']):
        self.circuit = circuit
        self.target = None
        self.op_type = op_type
        self.params = ()

    def __call__(self, *args):
        self.params = args
        return self

    def __getitem__(self, args):
        self.target = args
        self.circuit.ops.append(
            self.op_type.create(self.target, self.params))
        # ad-hoc
        self.circuit.n_qubits = max(
            gate.get_maximum_index(args) + 1, self.circuit.n_qubits)
        return self.circuit

    def __str__(self):
        if self.params:
            args_str = str(self.params)
        else:
            args_str = ""
        return self.op_type.lowername + args_str + " " + str(self.target)


class BlueqatGlobalSetting:
    """Setting for Blueqat."""
    @staticmethod
    def register_macro(name: str,
                       func: Callable,
                       allow_overwrite: bool = False) -> None:
        """Register new macro to Circuit.

        Args:
            | name (str): The name of macro.
            | func (callable): The function to be called.
            | allow_overwrite (bool, optional): If True, allow to overwrite the existing macro.
            |     Otherwise, raise the ValueError.

        Raises:
            ValueError: The name is duplicated with existing macro, gate or method.
                When `allow_overwrite=True`, this error is not raised.
        """
        if hasattr(Circuit, name):
            if allow_overwrite:
                warnings.warn(f"Circuit has attribute `{name}`.")
            else:
                raise ValueError(f"Circuit has attribute `{name}`.")
        if name.startswith("run_with_"):
            if allow_overwrite:
                warnings.warn(
                    f"Gate name `{name}` may conflict with run of backend.")
            else:
                raise ValueError(
                    f"Gate name `{name}` shall not start with 'run_with_'.")
        if not allow_overwrite:
            if get_op_type(name) is not None:
                raise ValueError(
                    f"Gate '{name}' is already exists in gate set.")
            if name in GLOBAL_MACROS:
                raise ValueError(f"Macro '{name}' is already exists.")
        GLOBAL_MACROS[name] = func

    @staticmethod
    def unregister_macro(name):
        """Unregister a macro.

        Args:
            name (str): The name of the macro to be unregistered.

        Raises:
            ValueError: Specified gate is not registered.
        """
        if name not in GLOBAL_MACROS:
            raise ValueError(f"Macro '{name}' is not registered.")
        del GLOBAL_MACROS[name]

    @staticmethod
    def register_gate(name, gateclass, allow_overwrite=False):
        """Register new gate to gate set.

        Args:
            | name (str): The name of gate.
            | gateclass (type): The type object of gate.
            | allow_overwrite (bool, optional): If True, allow to overwrite the existing gate.
            |     Otherwise, raise the ValueError.

        Raises:
            ValueError: The name is duplicated with existing gate.
                When `allow_overwrite=True`, this error is not raised.
        """
        if hasattr(Circuit, name):
            if allow_overwrite:
                warnings.warn(f"Circuit has attribute `{name}`.")
            else:
                raise ValueError(f"Circuit has attribute `{name}`.")
        if name.startswith("run_with_"):
            if allow_overwrite:
                warnings.warn(
                    f"Gate name `{name}` may conflict with run of backend.")
            else:
                raise ValueError(
                    f"Gate name `{name}` shall not start with 'run_with_'.")
        if not allow_overwrite:
            if get_op_type(name) is not None:
                raise ValueError(
                    f"Gate '{name}' is already exists in gate set.")
            if name in GLOBAL_MACROS:
                raise ValueError(f"Macro '{name}' is already exists.")
        register_operation(name, gateclass)

    @staticmethod
    def unregister_gate(name):
        """Unregister a gate from gate set

        Args:
            name (str): The name of the gate to be unregistered.

        Raises:
            ValueError: Specified gate is not registered.
        """
        if get_op_type(name) is None:
            raise ValueError(f"Gate '{name}' is not registered.")
        unregister_operation(name)

    @staticmethod
    def register_backend(name, backend, allow_overwrite=False):
        """Register new backend.

        Args:
            | name (str): The name of backend.
            | gateclass (type): The type object of backend
            | allow_overwrite (bool, optional): If True, allow to overwrite the existing backend.
                Otherwise, raise the ValueError.

        Raises:
            ValueError: The name is duplicated with existing backend.
                When `allow_overwrite=True`, this error is not raised.
        """
        if hasattr(Circuit, "run_with_" + name):
            if allow_overwrite:
                warnings.warn(f"Circuit has attribute `run_with_{name}`.")
            else:
                raise ValueError(f"Circuit has attribute `run_with_{name}`.")
        if not allow_overwrite:
            if name in BACKENDS:
                raise ValueError(
                    f"Backend '{name}' is already registered as backend.")
        BACKENDS[name] = backend

    @staticmethod
    def unregister_backend(name):
        """Unregister a backend.

        Args:
            name (str): The name of the backend to be unregistered.

        Raises:
            ValueError: Specified backend is not registered.
        """
        if name not in BACKENDS:
            raise ValueError(f"Backend '{name}' is not registered.")
        del BACKENDS[name]

    @staticmethod
    def remove_backend(name):
        """This method is deperecated. Use `unregister_backend` method."""
        warnings.warn(
            "remove_backend is deprecated. `unregister_backend` is recommended.",
            DeprecationWarning)
        BlueqatGlobalSetting.unregister_backend(name)

    @staticmethod
    def set_default_backend(name):
        """Set the default backend to be used by `Circuit`.
        Args:
            name (str): The name of the default backend.

        Raises:
            ValueError: Specified backend is not registered.
        """
        if name not in BACKENDS:
            raise ValueError(f"Backend '{name}' is not registered.")
        global DEFAULT_BACKEND_NAME
        DEFAULT_BACKEND_NAME = name

    @staticmethod
    def get_default_backend_name():
        """Get the default backend name.

        Returns:
            str: The name of default backend.
        """
        return DEFAULT_BACKEND_NAME


from .backends import BACKENDS, DEFAULT_BACKEND_NAME



# Copyright 2019 The Blueqat Developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The module for calculate Pauli matrices."""

from bisect import bisect_left
from collections import defaultdict, namedtuple
from functools import reduce
from itertools import combinations, groupby, product
from numbers import Number, Integral
from math import pi

import numpy as np
import scipy.sparse

_PauliTuple = namedtuple("_PauliTuple", "n")
half_pi = pi / 2

_sparse_types = {
    'bsr': scipy.sparse.bsr_matrix,
    'coo': scipy.sparse.coo_matrix,
    'csc': scipy.sparse.csc_matrix,
    'csr': scipy.sparse.csr_matrix,
    'dia': scipy.sparse.dia_matrix,
    'dok': scipy.sparse.dok_matrix,
    'lil': scipy.sparse.lil_matrix,
}

_matrix = {
    'I': np.array([[1, 0], [0, 1]], dtype=complex),
    'X': np.array([[0, 1], [1, 0]], dtype=complex),
    'Y': np.array([[0, -1j], [1j, 0]], dtype=complex),
    'Z': np.array([[1, 0], [0, -1]], dtype=complex)
}

_mul_map = {
    ('X', 'X'): (1.0, 'I'),
    ('X', 'Y'): (1j, 'Z'),
    ('X', 'Z'): (-1j, 'Y'),
    ('Y', 'X'): (-1j, 'Z'),
    ('Y', 'Y'): (1.0, 'I'),
    ('Y', 'Z'): (1j, 'X'),
    ('Z', 'X'): (1j, 'Y'),
    ('Z', 'Y'): (-1j, 'X'),
    ('Z', 'Z'): (1.0, 'I'),
}

_sparse_matrix = {
    ty: {ch: fn(mat, dtype=complex)
         for ch, mat in _matrix.items()}
    for ty, fn in _sparse_types.items()
}


def _kron_1d(a, b):
    """This function is for internal use.
    Returns a⊗b for 1d array.
    """
    nb = b.size
    d = np.repeat(a, nb).reshape(-1, nb)
    d *= b
    return d.reshape(-1)


def _kron_1d_rec(krons, cumsum, lo, hi):
    """This function is for internal use.
    Equivalent with reduce(_kron_1d, krons[lo:hi]), but faster.
    """
    if hi - lo == 1:
        return krons[lo]
    if hi - lo == 2:
        return _kron_1d(krons[lo], krons[lo + 1])
    mid = bisect_left(cumsum, (cumsum[lo] + cumsum[hi - 1]) // 2, lo, hi)
    if mid == lo:
        return _kron_1d(krons[lo], _kron_1d_rec(krons, cumsum, lo + 1, hi))
    return _kron_1d(_kron_1d_rec(krons, cumsum, lo, mid),
                    _kron_1d_rec(krons, cumsum, mid, hi))


def _term_to_dataarray(term, n_qubits, rowmajor):
    """This function is for internal use.
    Make data of sparse Kronecker product matrix."""
    y_mat = np.array([-1j, 1j]) if rowmajor else np.array([1j, -1j])
    paulis = ['I'] * n_qubits
    data_list = []
    for op in term.ops:
        paulis[op.n] = op.op
    for g, l in groupby(paulis):
        n = len(tuple(l))
        if g == 'Y':
            data_list += [y_mat.copy() for _ in range(n)]
        elif g == 'Z':
            data_list += [np.array([1, -1], dtype=complex) for _ in range(n)]
        else:
            data_list.append(np.repeat(np.array([1], dtype=complex), 2**n))
    t = min(data_list, key=len)
    t *= term.coeff
    data_list.reverse()
    cumsum = np.array([k.size for k in data_list]).cumsum()
    return _kron_1d_rec(data_list, cumsum, 0, len(cumsum))


def _term_to_indices(term, dim, dtype, rowcol):
    """This function is for internal use.
    Make indices for sparse Kronecker product matrix."""
    xor_bits = sum(1 << op.n for op in term.ops if op.op in 'XY')
    if rowcol:
        col = np.arange(dim, dtype=dtype)
        row = col ^ xor_bits
        return row, col
    return np.arange(dim, dtype=dtype) ^ xor_bits


def pauli_from_char(ch, n=0):
    """Make Pauli matrix from an character.

    Args:
        | ch (str): "X" or "Y" or "Z" or "I".
        | n (int, optional): Make Pauli matrix as n-th qubits.

    Returns:
        If ch is "X" => X, "Y" => Y, "Z" => Z, "I" => I

    Raises:
        ValueError: When ch is not "X", "Y", "Z" nor "I".
    """
    ch = ch.upper()
    if ch == "I":
        return I
    if ch == "X":
        return X(n)
    if ch == "Y":
        return Y(n)
    if ch == "Z":
        return Z(n)
    raise ValueError("ch shall be X, Y, Z or I")


def term_from_chars(chars):
    """Make Pauli's Term from chars which is written by "X", "Y", "Z" or "I".
    e.g. "XZIY" => X(3) * Z(2) * Y(0)

    Args:
        chars (str): Written in "X", "Y", "Z" or "I".

    Returns:
        Term: A `Term` object.

    Raises:
        ValueError: When chars conteins the character which is "X", "Y", "Z" nor "I".
    """
    return Term.from_chars(reversed(chars))


def to_term(pauli):
    """Convert to Term from Pauli operator (X, Y, Z, I).

    Args:
        pauli (X, Y, Z or I): A Pauli operator

    Returns:
        Term: A `Term` object.
    """
    return pauli.to_term()


def to_expr(term):
    """Convert to Expr from Term or Pauli operator (X, Y, Z, I).

    Args:
        term: (Term, X, Y, Z or I): A Term or Pauli operator.

    Returns:
        Expr: An `Expr` object.
    """
    return term.to_expr()


def commutator(expr1, expr2):
    """Returns [expr1, expr2] = expr1 * expr2 - expr2 * expr1.

    Args:
        | expr1 (Expr, Term or Pauli operator): Pauli's expression.
        | expr2 (Expr, Term or Pauli operator): Pauli's expression.

    Returns:
        Expr: expr1 * expr2 - expr2 * expr1.
    """
    expr1 = expr1.to_expr().simplify()
    expr2 = expr2.to_expr().simplify()
    return (expr1 * expr2 - expr2 * expr1).simplify()


def is_commutable(expr1, expr2, eps=0.00000001):
    """Test whether expr1 and expr2 are commutable.

    Args:
        | expr1 (Expr, Term or Pauli operator): Pauli's expression.
        | expr2 (Expr, Term or Pauli operator): Pauli's expression.
        | eps (float, optional): Machine epsilon.
            If | \[expr1, expr2 \]| < eps, consider it is commutable.

    Returns:
        bool: if expr1 and expr2 are commutable, returns True, otherwise False.
    """
    return sum((x * x.conjugate()).real
               for x in commutator(expr1, expr2).coeffs()) < eps


# To avoid pylint error
def _n(pauli):
    return pauli.n


def _GetItem(self_, n):
    return type(self_)(n)


class _PauliImpl:
    @property
    def op(self):
        """Return operator type (X, Y, Z, I)"""
        return self.__class__.__name__[1]

    @property
    def is_identity(self):
        """If `self` is I, returns True, otherwise False."""
        return self.op == "I"

    @property
    def n_qubits(self):
        """Returns `self.n + 1` if self is not I. otherwise 0."""
        return 0 if self.is_identity else _n(self) + 1

    def __hash__(self):
        return hash((self.op, _n(self)))

    def __eq__(self, other):
        if isinstance(other, _PauliImpl):
            if self.is_identity:
                return other.is_identity
            return _n(self) == _n(other) and self.op == other.op
        if isinstance(other, Term):
            return self.to_term() == other
        if isinstance(other, Expr):
            return self.to_expr() == other
        return NotImplemented

    def __ne__(self, other):
        return not self == other

    def __mul__(self, other):
        if isinstance(other, Number):
            return Term.from_pauli(self, other)
        if not isinstance(other, _PauliImpl):
            return NotImplemented
        if self.is_identity:
            return other.to_term()
        if other.is_identity:
            return self.to_term()
        if _n(self) == _n(other) and self.op == other.op:
            return I.to_term()
        return Term.from_paulipair(self, other)

    def __rmul__(self, other):
        if isinstance(other, Number):
            return Term.from_pauli(self, other)
        return NotImplemented

    def __truediv__(self, other):
        if isinstance(other, Number):
            if other:
                return Term.from_pauli(self, 1.0 / other)
            raise ZeroDivisionError
        return NotImplemented

    def __add__(self, other):
        return self.to_expr() + other

    def __radd__(self, other):
        return other + self.to_expr()

    def __sub__(self, other):
        return self.to_expr() - other

    def __rsub__(self, other):
        return other - self.to_expr()

    def __neg__(self):
        return Term.from_pauli(self, -1.0)

    def __repr__(self):
        if self.is_identity:
            return "I"
        return self.op + "[" + str(_n(self)) + "]"

    def to_term(self):
        """Convert to Pauli Term"""
        return Term.from_pauli(self)

    def to_expr(self):
        """Convert to Pauli Expr"""
        return self.to_term().to_expr()

    @property
    def matrix(self):
        """Matrix reprentation of this operator."""
        return _matrix[self.op].copy()

    def to_matrix(self, n_qubits=-1, *, sparse=None):
        """Convert to the matrix."""
        return self.to_term().to_matrix(n_qubits, sparse=sparse)


class _X(_PauliImpl, _PauliTuple):
    """Pauli's X operator"""


class _Y(_PauliImpl, _PauliTuple):
    """Pauli's Y operator"""


class _Z(_PauliImpl, _PauliTuple):
    """Pauli's Z operator"""


class _PauliCtor:
    def __init__(self, ty):
        self.ty = ty

    def __call__(self, n):
        return self.ty(n)

    def __getitem__(self, n):
        return self.ty(n)

    @property
    def matrix(self):
        """Matrix reprentation of this operator."""
        return _matrix[self.ty.__name__[-1]].copy()


X = _PauliCtor(_X)
Y = _PauliCtor(_Y)
Z = _PauliCtor(_Z)


class _I(_PauliImpl, namedtuple("_I", "")):
    """Identity operator"""
    def __call__(self):
        return self

    @property
    def matrix(self):
        """Matrix reprentation of this operator."""
        return _matrix['I'].copy()


I = _I()
_TermTuple = namedtuple("_TermTuple", "ops coeff")


class Term(_TermTuple):
    """Multiplication of Pauli matrices with coefficient.
    Note that this class is immutable.

    Multiplied Pauli matrices are very important for quantum computation
    because it is an unitary matrix (without coefficient) and also
    it can be consider the time evolution of the term (with real coefficient)
    without Suzuki-Trotter expansion.
    """
    @staticmethod
    def from_paulipair(pauli1, pauli2):
        """Make new Term from two Pauli operator."""
        return Term(Term.join_ops((pauli1, ), (pauli2, )), 1.0)

    @staticmethod
    def from_pauli(pauli, coeff=1.0):
        """Make new Term from an Pauli operator"""
        if pauli.is_identity or coeff == 0:
            return Term((), coeff)
        return Term((pauli, ), coeff)

    @staticmethod
    def from_ops_iter(ops, coeff):
        """For internal use."""
        return Term(tuple(ops), coeff)

    @staticmethod
    def from_chars(chars):
        """Make Pauli's Term from chars which is written by "X", "Y", "Z" or "I".
        e.g. "XZIY" => X(0) * Z(1) * Y(3)

        Args:
            chars (str): Written in "X", "Y", "Z" or "I".

        Returns:
            Term: A `Term` object.

        Raises:
            ValueError: When chars conteins the character which is "X", "Y", "Z" nor "I".
        """
        paulis = [
            pauli_from_char(c, n) for n, c in enumerate(chars) if c != "I"
        ]
        if not paulis:
            return 1.0 * I
        if len(paulis) == 1:
            return 1.0 * paulis[0]
        return reduce(lambda a, b: a * b, paulis)

    @staticmethod
    def join_ops(ops1, ops2):
        """For internal use."""
        i = len(ops1) - 1
        j = 0
        while i >= 0 and j < len(ops2):
            if ops1[i] == ops2[j]:
                i -= 1
                j += 1
            else:
                break
        return ops1[:i + 1] + ops2[j:]

    @property
    def is_identity(self):
        """If `self` is I, returns True, otherwise False."""
        return not self.ops

    def __mul__(self, other):
        if isinstance(other, Number):
            return Term(self.ops, self.coeff * other)
        if isinstance(other, Term):
            ops = Term.join_ops(self.ops, other.ops)
            coeff = self.coeff * other.coeff
            return Term(ops, coeff)
        if isinstance(other, _PauliImpl):
            if other.is_identity:
                return self
            return Term(Term.join_ops(self.ops, (other, )), self.coeff)
        return NotImplemented

    def __rmul__(self, other):
        if isinstance(other, Number):
            return Term(self.ops, self.coeff * other)
        if isinstance(other, _PauliImpl):
            if other.is_identity:
                return self
            return Term(Term.join_ops((other, ), self.ops), self.coeff)
        return NotImplemented

    def __truediv__(self, other):
        if isinstance(other, (int, float)):
            if other:
                return Term(self.ops, self.coeff / other)
            raise ZeroDivisionError
        return NotImplemented

    def __pow__(self, n):
        if isinstance(n, Integral):
            if n < 0:
                raise ValueError(
                    "`pauli_term ** n` or `pow(pauli_term, n)`: " +
                    "n shall not be negative value.")
            if n == 0:
                return Term.from_pauli(I)
            return Term(self.ops * n, self.coeff**n)
        return NotImplemented

    def __add__(self, other):
        return Expr.from_term(self) + other

    def __radd__(self, other):
        return other + Expr.from_term(self)

    def __sub__(self, other):
        return Expr.from_term(self) - other

    def __rsub__(self, other):
        return other - Expr.from_term(self)

    def __neg__(self):
        return Term(self.ops, -self.coeff)

    def __repr__(self):
        if self.coeff == 0:
            return "0*I"
        if self.coeff == -1.0:
            s_coeff = "-"
        else:
            s_coeff = str(self.coeff) + "*"
        if self.ops == ():
            s_ops = "I"
        else:
            s_ops = "*".join(op.op + "[" + repr(op.n) + "]" for op in self.ops)
        return s_coeff + s_ops

    def __eq__(self, other):
        if isinstance(other, _PauliImpl):
            other = other.to_term()
        return _TermTuple.__eq__(self, other) or \
               _TermTuple.__eq__(self.simplify(), other.simplify())

    def __ne__(self, other):
        return not self == other

    def to_term(self):
        """Do nothing. This method is prepared to avoid TypeError."""
        return self

    def to_expr(self):
        """Convert to Expr."""
        return Expr.from_term(self)

    def commutator(self, other):
        """Returns commutator."""
        return commutator(self, other)

    def is_commutable_with(self, other):
        """Test whether `self` is commutable with `other`."""
        return is_commutable(self, other)

    def simplify(self):
        """Simplify the Term."""
        def mul(op1, op2):
            if op1 == "I":
                return 1.0, op2
            if op2 == "I":
                return 1.0, op1
            return _mul_map[op1, op2]

        before = defaultdict(list)
        for op in self.ops:
            if op.op == "I":
                continue
            before[op.n].append(op.op)
        new_coeff = self.coeff
        new_ops = []
        for n in sorted(before.keys()):
            ops = before[n]
            assert ops
            k = 1.0
            op = ops[0]
            for _op in ops[1:]:
                _k, op = mul(op, _op)
                k *= _k
            new_coeff *= k
            if new_coeff.imag == 0:
                # cast to float
                new_coeff = new_coeff.real
            if op != "I":
                new_ops.append(pauli_from_char(op, n))
        return Term(tuple(new_ops), new_coeff)

    def n_iter(self):
        """Returns an iterator which yields indices for each Pauli matrices in the Term."""
        return (op.n for op in self.ops)

    def max_n(self):
        """Returns the maximum index of Pauli matrices in the Term.
        If there's no Pauli matrices, returns -1.
        """
        try:
            return max(self.n_iter())
        except ValueError:
            return -1

    @property
    def n_qubits(self):
        """Returns the number of qubits of the term.
        If the term is constant with identity matrix, n_qubits is 0."""
        return self.max_n() + 1

    def append_to_circuit(self, circuit, simplify=True):
        """Append Pauli gates to `Circuit`."""
        if simplify:
            term = self.simplify()
        else:
            term = self
        for op in term.ops[::-1]:
            gate = op.op.lower()
            if gate != "i":
                getattr(circuit, gate)[op.n]

    def get_time_evolution(self):
        """Get the function to append the time evolution of this term.

        Returns:
            function(circuit: Circuit, t: float):
                Add gates for time evolution to `circuit` with time `t`
        """
        term = self.simplify()
        coeff = term.coeff
        if coeff.imag:
            raise ValueError("Not a real coefficient.")
        ops = term.ops

        def append_to_circuit(circuit, t):
            if not ops:
                return
            for op in ops:
                n = op.n
                if op.op == "X":
                    circuit.h[n]
                elif op.op == "Y":
                    circuit.rx(-half_pi)[n]
            for i in range(1, len(ops)):
                circuit.cx[ops[i - 1].n, ops[i].n]
            circuit.rz(-2 * coeff * t)[ops[-1].n]
            for i in range(len(ops) - 1, 0, -1):
                circuit.cx[ops[i - 1].n, ops[i].n]
            for op in ops:
                n = op.n
                if op.op == "X":
                    circuit.h[n]
                elif op.op == "Y":
                    circuit.rx(half_pi)[n]

        return append_to_circuit

    def to_matrix(self, n_qubits=-1, *, sparse=None):
        """Convert to the matrix."""
        if not (sparse is None or sparse in _sparse_types):
            raise ValueError(f'Unknown sparse format {sparse}.')
        if n_qubits == -1:
            n_qubits = self.n_qubits
        if n_qubits == 0:
            m = np.array([[self.coeff]])
            if sparse is None:
                return m
            return _sparse_types[sparse](m)
        dim = 2**n_qubits
        term = self.simplify()
        data = _term_to_dataarray(term, n_qubits, sparse == 'csr')
        dtype_idx = np.int32 if n_qubits < 31 else np.int64
        if sparse == 'csc':
            indices = _term_to_indices(term, dim, dtype_idx, False)
            return scipy.sparse.csc_matrix(
                (data, indices, np.arange(dim + 1, dtype=dtype_idx)),
                shape=(dim, dim))
        if sparse == 'csr':
            indices = _term_to_indices(term, dim, dtype_idx, False)
            return scipy.sparse.csr_matrix(
                (data, indices, np.arange(dim + 1, dtype=dtype_idx)),
                shape=(dim, dim))
        row, col = _term_to_indices(term, dim, dtype_idx, True)
        m = scipy.sparse.coo_matrix((data, (row, col)), shape=(dim, dim))
        if sparse is None:
            return m.toarray()
        return _sparse_types[sparse](m)


_ExprTuple = namedtuple("_ExprTuple", "terms")


class Expr(_ExprTuple):
    @staticmethod
    def from_number(num):
        """Make new Expr from a number"""
        if num:
            return Expr.from_term(Term((), num))
        return Expr.zero()

    @staticmethod
    def from_term(term):
        """Make new Expr from a Term"""
        if term.coeff:
            return Expr((term, ))
        return Expr.zero()

    @staticmethod
    def from_terms_iter(terms):
        """For internal use."""
        return Expr(tuple(term for term in terms if term.coeff))

    def terms_to_dict(self):
        """For internal use."""
        return {term[0]: term[1] for term in self.terms if term.coeff}

    @staticmethod
    def from_terms_dict(terms_dict):
        """For internal use."""
        return Expr(tuple(Term(k, v) for k, v in terms_dict.items() if v))

    @staticmethod
    def zero():
        """Returns 0 as Term"""
        return Expr(())

    @property
    def is_identity(self):
        """If `self` is I, returns True, otherwise False."""
        if not self.terms:
            return True
        return len(
            self.terms
        ) == 1 and not self.terms[0].ops and self.terms[0].coeff == 1.0

    def __eq__(self, other):
        if isinstance(other, (_PauliImpl, Term)):
            other = other.to_expr()
        if isinstance(other, Expr):
            return self.terms == other.terms or self.simplify(
            ).terms == other.simplify().terms
        return NotImplemented

    def __ne__(self, other):
        return not self == other

    def __add__(self, other):
        if isinstance(other, Number):
            other = Expr.from_number(other)
        elif isinstance(other, Term):
            other = Expr.from_term(other)
        if isinstance(other, Expr):
            terms = self.terms_to_dict()
            for op, coeff in other.terms:
                if op in terms:
                    terms[op] += coeff
                    if terms[op] == 0:
                        del terms[op]
                else:
                    terms[op] = coeff
            return Expr.from_terms_dict(terms)
        return NotImplemented

    def __sub__(self, other):
        if isinstance(other, Number):
            other = Expr.from_number(other)
        elif isinstance(other, Term):
            other = Expr.from_term(other)
        if isinstance(other, Expr):
            terms = self.terms_to_dict()
            for op, coeff in other.terms:
                if op in terms:
                    terms[op] -= coeff
                    if terms[op] == 0:
                        del terms[op]
                else:
                    terms[op] = -coeff
            return Expr.from_terms_dict(terms)
        return NotImplemented

    def __radd__(self, other):
        if isinstance(other, Number):
            return Expr.from_number(other) + self
        if isinstance(other, Term):
            return Expr.from_term(other) + self
        return NotImplemented

    def __rsub__(self, other):
        if isinstance(other, Number):
            return Expr.from_number(other) - self
        if isinstance(other, Term):
            return Expr.from_term(other) - self
        return NotImplemented

    def __neg__(self):
        return Expr(tuple(Term(op, -coeff) for op, coeff in self.terms))

    def __mul__(self, other):
        if isinstance(other, Number):
            if other == 0:
                return Expr.from_number(0.0)
            return Expr.from_terms_iter(
                Term(op, coeff * other) for op, coeff in self.terms)
        if isinstance(other, _PauliImpl):
            other = other.to_term()
        if isinstance(other, Term):
            return Expr(tuple(term * other for term in self.terms))
        if isinstance(other, Expr):
            terms = defaultdict(float)
            for t1, t2 in product(self.terms, other.terms):
                term = t1 * t2
                terms[term.ops] += term.coeff
            return Expr.from_terms_dict(terms)
        return NotImplemented

    def __rmul__(self, other):
        if isinstance(other, Number):
            if other == 0:
                return Expr.from_number(0.0)
            return Expr.from_terms_iter(
                Term(op, coeff * other) for op, coeff in self.terms)
        if isinstance(other, _PauliImpl):
            other = other.to_term()
        if isinstance(other, Term):
            return Expr(tuple(other * term for term in self.terms))
        return NotImplemented

    def __truediv__(self, other):
        if isinstance(other, Number):
            if other:
                return Expr(tuple(term / other for term in self.terms))
            raise ZeroDivisionError
        return NotImplemented

    def __pow__(self, n):
        if isinstance(n, Integral):
            if n < 0:
                raise ValueError(
                    "`pauli_expr ** n` or `pow(pauli_expr, n)`: " +
                    "n shall not be negative value.")
            if n == 0:
                return Expr.from_number(1.0)
            val = self
            for _ in range(n - 1):
                val *= self
            return val
        return NotImplemented

    def __iter__(self):
        return iter(self.terms)

    def __repr__(self):
        if not self.terms:
            return "0*I+0"
        s_terms = [repr(self.terms[0])]
        for term in self.terms[1:]:
            s = repr(term)
            if s[0] == "+":
                s_terms.append("+")
                s_terms.append(s[1:])
            elif s[0] == "-":
                s_terms.append("-")
                s_terms.append(s[1:])
            else:
                s_terms.append("+")
                s_terms.append(s)
        return " ".join(s_terms)

    def __getnewargs__(self):
        return (self.terms, )

    def to_expr(self):
        """Do nothing. This method is prepared to avoid TypeError."""
        return self

    def max_n(self):
        """Returns the maximum index of Pauli matrices in the Expr.
        If Expr is empty or only constant and identity matrix, returns -1.
        """
        try:
            return max(term.max_n() for term in self.terms if term.ops)
        except ValueError:
            return -1

    @property
    def n_qubits(self):
        """Returns the number of qubits of the Term.

        If Expr is empty or only constant and identity matrix, returns 0.
        """
        return self.max_n() + 1

    def coeffs(self):
        """Generator which yields a coefficent for each Term."""
        for term in self.terms:
            yield term.coeff

    def commutator(self, other):
        """Returns commutator."""
        return commutator(self, other)

    def is_commutable_with(self, other):
        """Test whether `self` is commutable with `other`."""
        return is_commutable(self, other)

    def is_all_terms_commutable(self):
        """Test whether all terms are commutable. This function may very slow."""
        return all(is_commutable(a, b) for a, b in combinations(self.terms, 2))

    def simplify(self):
        """Simplify the Expr."""
        d = defaultdict(float)
        for term in self.terms:
            term = term.simplify()
            d[term.ops] += term.coeff
        return Expr.from_terms_iter(
            Term.from_ops_iter(k, d[k]) for k in sorted(d, key=repr) if d[k])

    def to_matrix(self, n_qubits=-1, *, sparse=None):
        """Convert to the matrix."""
        if not (sparse is None or sparse in _sparse_types):
            raise ValueError(f'Unknown sparse format {sparse}.')
        if n_qubits == -1:
            n_qubits = self.n_qubits
        if n_qubits == 0:
            m = np.array([[sum(term.coeff for term in self.terms)]])
            if sparse is None:
                return m
            return _sparse_types[sparse](m)
        expr = self.simplify()
        grpkey = lambda pau: sum(1 << op.n for op in pau.ops if op.op in 'XY')
        dim = 2**n_qubits
        is_csr = sparse == 'csr'
        gr_terms = [
            list(g)
            for _, g in groupby(sorted(expr.terms, key=grpkey), key=grpkey)
        ]
        n_groups = len(gr_terms)
        n_vals = n_groups * dim
        dtype_idx = np.int32 if n_qubits < 31 else np.int64
        vals = np.empty(n_vals, dtype=complex)
        inds = np.empty(n_vals, dtype=dtype_idx)
        for i_grp, grp in enumerate(gr_terms):
            val_acc = _term_to_dataarray(grp[0], n_qubits, is_csr)
            inds[i_grp::n_groups] = _term_to_indices(grp[0], dim, dtype_idx,
                                                     False)
            for term in grp[1:]:
                val_acc += _term_to_dataarray(term, n_qubits, is_csr)
            vals[i_grp::n_groups] = val_acc
        if not is_csr:
            m = scipy.sparse.csc_matrix(
                (vals, inds, np.arange(0, n_vals + 1, n_groups)),
                shape=(dim, dim))
        else:
            m = scipy.sparse.csr_matrix(
                (vals, inds, np.arange(0, n_vals + 1, n_groups)),
                shape=(dim, dim))
        m.eliminate_zeros()
        if sparse is None:
            return m.toarray()
        return _sparse_types[sparse](m)


def qubo_bit(n):
    """Represent QUBO's bit to Pauli operator of Ising model.

    Args:
        n (int): n-th bit in QUBO

    Returns:
        Expr: Pauli expression of QUBO bit.
    """
    return 0.5 - 0.5 * Z[n]



# Copyright 2019 The Blueqat Developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module provides IBM Q Backend via OpenQASM and Qiskit.

To run this backend, call
Circuit.run_with_ibmq or Circuit.run(backend="ibmq", ...).

Args:
    qiskit_backend (Backend, optional): Backend of qiskit.
        If this argument is omitted, Blueqat uses Qiskit Aer's qasm_simulator.
    shots (int, optional): The number of shots.
    returns (str, optional): Choose from followings. Default is "shots".
        "shots": Default behavior. Returns the measurement result.
        "draw": Draw the circuit via `qiskit.circuit.QuantumCircuit.draw()`.
        "qiskit_circuit": Returns Qiskit's circuit class. (qiskit.circuit.QuantumCircuit)
        "qiskit_job": Returns Qiskit's job class. (qiskit.providers.BaseJob)
        "qiskit_result": Returns Qiskit's result class, which is returns by BaseJob.result().
        "_exception": This is for troubleshooting purpose.
                      Returns the exception which raised during importing Qiskit.
                      If no exception is raised, returns None.
    kwargs: Other arguments which passed to qiskit.execute
Returns:
    The result of run.
"""

import warnings
from collections import Counter
from .qasm_parser_backend_generator import generate_backend


def _qasm_runner_qiskit(qasm,
                        qiskit_backend=None,
                        shots=None,
                        returns=None,
                        **kwargs):
    if returns is None:
        returns = "shots"
    elif returns not in ("shots", "draw", "_exception", "qiskit_circuit",
                         "qiskit_job", "qiskit_result"):
        raise ValueError(
            "`returns` shall be None, 'shots', 'draw', " +
            "'qiskit_circuit', 'qiskit_job', 'qiskit_result' or '_exception'")

    import_error = None
    try:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            from qiskit import Aer, QuantumCircuit, execute
    except Exception as e:
        import_error = e

    if import_error:
        if returns == "_exception":
            return import_error
        if isinstance(import_error, ImportError):
            raise ImportError(
                "Cannot import qiskit. To use this backend, please install qiskit."
                + " `pip install qiskit`.")
        else:
            raise ValueError(
                "Unknown error raised when importing qiskit. To get exception, "
                + 'run this backend with arg `returns="_exception"`')
    else:
        if returns == "_exception":
            return None
        qk_circuit = QuantumCircuit.from_qasm_str(qasm)
        if returns == "qiskit_circuit":
            return qk_circuit
        if returns == "draw":
            return qk_circuit.draw(**kwargs)
        if shots is None:
            shots = 1024
        if qiskit_backend is None:
            qiskit_backend = Aer.get_backend("qasm_simulator")
        job = execute(qk_circuit,
                      backend=qiskit_backend,
                      shots=shots,
                      **kwargs)
        if returns == "qiskit_job":
            return job
        result = job.result()
        if returns == "qiskit_result":
            return result
        counts = Counter(
            {bits[::-1]: val
             for bits, val in result.get_counts().items()})
        return counts


ibmq_backend = generate_backend(_qasm_runner_qiskit)



# Copyright 2019 The Blueqat Developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import numpy as np


def pauli(qubo):
    """
    Convert to pauli operators of universal gate model.
    """
    from blueqat.pauli import qubo_bit
    h = 0.0
    assert all(len(q) == len(qubo) for q in qubo)
    for i in range(len(qubo)):
        h += qubo_bit(i) * qubo[i][i]
        for j in range(i + 1, len(qubo)):
            h += qubo_bit(i) * qubo_bit(j) * (qubo[i][j] + qubo[j][i])
    return h


def optx(quboh):
    try:
        import sympy
    except ImportError:
        raise ImportError(
            "optx() requires sympy. Please install before call this function.")

    optx_E = sympy.expand(quboh)
    symbol_list = list(optx_E.free_symbols)
    sympy.var(' '.join(map(str, symbol_list)), positive=True)
    for i in range(len(symbol_list)):
        optx_E = optx_E.subs(symbol_list[i] * symbol_list[i], symbol_list[i])
    return optx_E


def optm(quboh, numM):
    try:
        import sympy
    except ImportError:
        raise ImportError(
            "optm() requires sympy. Please install before call this function.")
    optm_E = sympy.expand(quboh)
    symbol_list = ["q" + str(i) for i in range(numM)]
    sympy.var(' '.join(symbol_list), positive=True)

    symbol_list_proto = list(optm_E.free_symbols)
    for i in range(len(symbol_list_proto)):
        optm_E = optm_E.subs(symbol_list_proto[i] * symbol_list_proto[i],
                             symbol_list_proto[i])

    optm_M = np.zeros((numM, numM))

    for i in range(numM):
        for j in range(i + 1, numM):
            optm_M[i][j] = optm_E.coeff(symbol_list[i] + "*" + symbol_list[j])

        temp1 = sympy.poly(optm_E.coeff(symbol_list[i]))
        optm_M[i][i] = sympy.poly(optm_E.coeff(
            symbol_list[i])).coeff_monomial(1)

    return optm_M


def make_qs(n, m=None):
    """Make sympy symbols q0, q1, ...
    
    Args:
        n(int), m(int, optional):
            | If specified both n and m, returns [qn, q(n+1), ..., qm],
            | Only n is specified, returns[q0, q1, ..., qn].
    Return:
        tuple(Symbol): Tuple of sympy symbols.
    """
    try:
        import sympy
    except ImportError:
        raise ImportError("This function requires sympy. Please install it.")
    if m is None:
        syms = sympy.symbols(" ".join(f"q{i}" for i in range(n)))
        if isinstance(syms, tuple):
            return syms
        else:
            return (syms, )
    syms = sympy.symbols(" ".join(f"q{i}" for i in range(n, m)))
    if isinstance(syms, tuple):
        return syms
    else:
        return (syms, )


def nbody_separation(expr, qs):
    """Convert n-body problem to 2-body problem.
    
    Args:
        | expr: sympy expressions to be separated.
        | qs: sympy's symbols to be used as supplementary variable.
    Return:
        new_expr(sympy expr), constraints(sympy expr), mapping(dict(str, str -> Symbol)):
            | `new_expr` is converted problem, `constraints` is constraints for supplementary variable.
            | You may use `expr = new_expr + delta * constraints`, delta is floating point variable.
            | mapping is supplementary variable's mapping.
    """
    try:
        import sympy
    except ImportError:
        raise ImportError("This function requires sympy. Please install it.")
    logging.debug(expr)
    free_symbols = expr.free_symbols
    logging.debug(free_symbols)
    assert type(expr) == sympy.Add
    logging.debug(expr.args)
    mapping = {}
    new_expr = sympy.expand(0)
    constraints = sympy.expand(0)
    i_var = 0
    for arg in expr.args:
        if isinstance(arg, sympy.Symbol):
            new_expr += arg
            continue
        if not arg.free_symbols:
            new_expr += arg
            continue
        assert type(arg) == sympy.Mul
        syms = arg.free_symbols.copy()
        while len(syms) > 2:
            it = iter(syms)
            for v1, v2 in zip(it, it):
                if (str(v1), str(v2)) in mapping:
                    v = mapping[str(v1), str(v2)]
                    logging.debug(f"{v1}*{v2} -> {v} (Existed variable)")
                else:
                    v = qs[i_var]
                    i_var += 1
                    mapping[(str(v1), str(v2))] = v
                    logging.debug(f"{v1}*{v2} -> {v} (New variable)")
                    constraints += 3 * v + v1 * v2 - 2 * v1 * v - 2 * v2 * v
                    logging.debug(f"constraints: {constraints}")
                arg = arg.subs(v1 * v2, v)
            syms = arg.free_symbols.copy()
        new_expr += arg
        logging.debug(f"new_expr: {new_expr}")
    return new_expr, constraints, mapping


def qn_to_qubo(expr):
    """Convert Sympy's expr to QUBO.
    
    Args:
        expr: Sympy's quadratic expression with variable `q0`, `q1`, ...
    Returns:
        [[float]]: Returns QUBO matrix.
    """
    try:
        import sympy
    except ImportError:
        raise ImportError("This function requires sympy. Please install it.")
    assert type(expr) == sympy.Add
    to_i = lambda s: int(str(s)[1:])
    max_i = max(map(to_i, expr.free_symbols)) + 1
    qubo = [[0.] * max_i for _ in range(max_i)]
    for arg in expr.args:
        syms = arg.free_symbols
        assert len(syms) <= 2
        if len(syms) == 2:
            assert type(arg) == sympy.Mul
            i, j = list(map(to_i, syms))
            if i > j:
                i, j = j, i
            if i == j:
                if len(arg.args) == 2:
                    qubo[i][i] = float(arg.args[0])
                elif len(arg.args) == 1:
                    qubo[i][i] = 1.0
                else:
                    raise ValueError(f"Too many args! arg.args = {arg.args}")
                continue
            if len(arg.args) == 3:
                qubo[i][j] = float(arg.args[0])
            elif len(arg.args) == 2:
                qubo[i][j]
        if len(syms) == 1:
            if len(arg.args) == 2:
                assert type(arg) == sympy.Mul
                i = to_i(next(iter(syms)))
                qubo[i][i] = float(arg.args[0])
            elif len(arg.args) == 1:
                qubo[i][i] = 1.0
            else:
                raise ValueError(f"Too many args! arg.args = {arg.args}")
    return qubo


def Ei(q3, j3):
    EE = 0
    for i in range(len(q3)):
        EE += q3[i] * j3[i][i]
        EE += sum(q3[i] * q3[i + 1:] * j3[i][i + 1:])
    return EE


def Ei_sqa(q, J, T, P, G):
    E = 0
    for p in range(P):
        E += Ei(q[p], J) / P
        E += T / 2 * np.log(1 / np.tanh(G / T / P)) * np.sum(
            q[p, :] * q[(p + 1 + P) % P, :])
    return E


def sel(selN, selK, selarr=[]):
    """
    Automatically create QUBO which select K qubits from N qubits

    .. code-block:: python

        print(wq.sel(5,2))
        #=>
        [[-3  2  2  2  2]
        [ 0 -3  2  2  2]
        [ 0  0 -3  2  2]
        [ 0  0  0 -3  2]
        [ 0  0  0  0 -3]]
        
    if you set array on the 3rd params, the result likely to choose the nth qubit in the array

    .. code-block:: python

        print(wq.sel(5,2,[0,2]))
        #=>
        [[-3.5  2.   2.   2.   2. ]
        [ 0.  -3.   2.   2.   2. ]
        [ 0.   0.  -3.5  2.   2. ]
        [ 0.   0.   0.  -3.   2. ]
        [ 0.   0.   0.   0.  -3. ]]
    """
    selres = np.diag([1 - 2 * selK] * selN) + np.triu(
        [[2] * selN for i in range(selN)], k=1)
    selmat = np.zeros(selN)
    for i in range(len(selarr)):
        selmat[selarr[i]] += 1
    selres = np.asarray(selres) - 0.5 * np.diag(selmat)
    return selres


def mul(mulA, mulB):
    return np.triu(np.outer(mulA, mulB)) + np.triu(np.outer(mulA, mulB), k=1)


def sqr(sqrA):
    return np.triu(np.outer(sqrA, sqrA)) + np.triu(np.outer(sqrA, sqrA), k=1)


def net(narr, nnet):
    """
    Automatically create QUBO which has value 1 for all connectivity defined by array of edges and graph size N

    .. code-block:: python

        print(wq.net([[0,1],[1,2]],4))
        #=>
        [[0. 1. 0. 0.]
        [0. 0. 1. 0.]
        [0. 0. 0. 0.]
        [0. 0. 0. 0.]]

    this create 4*4 QUBO and put value 1 on connection between 0th and 1st qubit, 1st and 2nd qubit
    """
    mat = np.zeros((nnet, nnet))
    for i in range(len(narr)):
        narr[i] = np.sort(narr[i])
        mat[narr[i][0]][narr[i][1]] = 1
    return mat


def counter(narr):
    import collections
    dis = []
    for i in range(len(narr)):
        dis.append(''.join([str(x) for x in narr[i]]))
    return collections.Counter(dis)


def diag(diag_ele):
    """
    Create QUBO with diag from list

    .. code-block:: python

        print(wq.diag([1,2,1]))
        #=>
        [[1 0 0]
        [0 2 0]
        [0 0 1]]
    """
    return np.diag(diag_ele)


def zeros(zeros_ele):
    """
    Create QUBO with all element value as 0

    .. code-block:: python

        print(wq.zeros(3))
        #=>
        [[0. 0. 0.]
        [0. 0. 0.]
        [0. 0. 0.]]
    """
    return np.zeros((zeros_ele, zeros_ele))


def rands(rands_ele):
    """
    Create QUBO with random number

    .. code-block:: python

        print(wq.rands(2))
        #=>
        [[0.89903411 0.68839641]
        [0.         0.28554602]]
    """
    return np.triu(np.random.rand(rands_ele, rands_ele))


def dbms(dbms_arr, weight_init=0):
    dbmN = np.sum(dbms_arr)
    if weight_init == 0:
        dbms_mat = np.diag(np.random.rand(dbmN))
    else:
        dbms_mat = np.diag([0.5 for i in range(dbmN)])
    cc = 0
    for k in range(len(dbms_arr) - 1):
        for i in range(cc, cc + dbms_arr[k]):
            for j in range(cc + dbms_arr[k],
                           cc + dbms_arr[k] + dbms_arr[k + 1]):
                if weight_init == 0:
                    dbms_mat[i][j] = np.random.rand()
                else:
                    dbms_mat[i][j] = weight_init
        cc = cc + dbms_arr[k]
    return dbms_mat


def sigm(x):
    return 1.0 / (1.0 + np.exp(-x))


def rbm_hmodel(vdata, dbms_mat1):
    hmodel = []
    vdataN = len(vdata)
    dbmsN = len(dbms_mat1)
    for i in range(vdataN, dbmsN):
        hsum = dbms_mat1[i][i]
        for j in range(vdataN):
            hsum += dbms_mat1[j][i] * vdata[j]
        hmodel.append(sigm(hsum))
    return hmodel


def rbm_data_qubo(vdata, dbms_mat1):
    hdata = rbm_hmodel(vdata, dbms_mat1)
    data_qubo = np.diag(vdata + hdata)
    vdataN = len(vdata)
    dbmsN = len(dbms_mat1)
    for i in range(vdataN):
        for j in range(vdataN, dbmsN):
            data_qubo[i][j] = vdata[i] * hdata[j - vdataN]
    return data_qubo


class Opt:
    """
    Optimizer for SA/SQA.
    """
    def __init__(self):
        #: Initial temperature [SA]
        self.Ts = 5
        #: Final temperature [SA]. Temperature [SQA]
        self.Tf = 0.02

        #: Initial strength of transverse magnetic field. [SQA]
        self.Gs = 10
        #: Final strength of transverse magnetic field. [SQA]
        self.Gf = 0.02
        #: Trotter slices [SQA]
        self.tro = 8

        #: Descreasing rate of temperature [SA]
        self.R = 0.95
        #: Iterations [SA]
        self.ite = 1000
        #: QUBO
        self.qubo = []
        self.J = []

        self.ep = 0
        #: List of energies
        self.E = []

        self.dwaveendpoint = 'https://cloud.dwavesys.com/sapi'
        self.dwavetoken = ''
        self.dwavesolver = 'DW_2000Q_6'

        #: RBM Models
        self.RBMvisible = 0
        self.RBMhidden = 0

    def reJ(self):
        return np.triu(self.J) + np.triu(self.J, k=1).T

    def qi(self):
        nn = len(self.qubo)
        qubo = np.triu(self.qubo + np.transpose(self.qubo)) - np.eye(nn) * self.qubo
        self.J = [np.random.choice([1., 1.], nn) for j in range(nn)]
        for i in range(nn):
            for j in range(i + 1, nn):
                self.J[i][j] = qubo[i][j] / 4

        self.J = np.triu(self.J) + np.triu(self.J, k=1).T

        for i in range(nn):
            sum = 0
            for j in range(nn):
                if i == j:
                    sum += qubo[i][i] * 0.5
                else:
                    sum += self.J[i][j]
            self.J[i][i] = sum

        self.ep = 0

        for i in range(nn):
            self.ep += self.J[i][i]
            for j in range(i + 1, nn):
                self.ep -= self.J[i][j]

        self.J = np.triu(self.J)

    def plot(self):
        """
        Draws energy chart using matplotlib.
        """
        import matplotlib.pyplot as plt
        plt.plot(self.E)
        plt.show()

    def qubo_to_matrix(self, qubo):
        try:
            import sympy
        except ImportError:
            raise ImportError(
                "optm() requires sympy. Please install before call this function."
            )

        qubo = self.expand_qubo(qubo)
        numN = len(qubo.free_symbols)
        optm = np.zeros((numN, numN))

        for i in qubo.free_symbols:
            for j in qubo.free_symbols:
                if (i != j):
                    optm[int(repr(i)[1:])][int(repr(j)[1:])] = qubo.coeff(i *
                                                                          j)
                else:
                    f2 = sympy.re(qubo.coeff(i))
                    for k in qubo.free_symbols:
                        f2 = f2.subs(k, 0)
                    optm[int(repr(i)[1:])][int(repr(i)[1:])] = f2

        return np.triu(optm)

    def expand_qubo(self, qubo):
        try:
            import sympy
        except ImportError:
            raise ImportError(
                "optm() requires sympy. Please install before call this function."
            )
        f = sympy.expand(qubo)
        deg = sympy.poly(f).degree()

        for i in range(deg):
            for j in f.free_symbols:
                f = f.subs(j**(deg - i), j)
        return f

    def add(self, qubo, M=1):
        len1 = len(self.qubo)
        if (isinstance(qubo, str)):
            qubo = self.qubo_to_matrix(self.expand_qubo(qubo))
        len2 = len(qubo)
        if (len1 == len2):
            self.qubo = np.array(self.qubo) + M * np.array(qubo)
        elif (self.qubo == []):
            self.qubo = M * np.array(qubo)
        return self

    def run(self, shots=1, sampler="normal", targetT=0.02, verbose=False):
        """
        Run SA with provided QUBO. 
        Set qubo attribute in advance of calling this method.
        """
        if self.qubo != []:
            self.qi()
        J = self.reJ()
        N = len(J)

        if sampler == "fast":
            itetemp = 100
            Rtemp = 0.75
        else:
            itetemp = self.ite
            Rtemp = self.R

        self.E = []
        qq = []
        for i in range(shots):
            T = self.Ts
            q = np.random.choice([-1, 1], N)
            EE = []
            EE.append(Ei(q, self.J) + self.ep)
            while T > targetT:
                x_list = np.random.randint(0, N, itetemp)
                for x in x_list:
                    q2 = np.ones(N) * q[x]
                    q2[x] = 1
                    dE = -2 * np.sum(q * q2 * J[:, x])

                    if dE < 0 or np.exp(-dE / T) > np.random.rand():
                        q[x] *= -1
                EE.append(Ei(q, self.J) + self.ep)
                T *= Rtemp
            self.E.append(EE)
            qtemp = (np.asarray(q, int) + 1) / 2
            qq.append([int(s) for s in qtemp])
            if verbose:
                print(i, ':', [int(s) for s in qtemp])
            if shots == 1:
                qq = qq[0]
        if shots == 1:
            self.E = self.E[0]
        return qq

    def sa(self, shots=1, sampler="normal"):
        sar = self.run(shots=shots, sampler=sampler)
        return sar

    def sqa(self):
        """
        Run SQA with provided QUBO.
        Set qubo attribute in advance of calling this method.
        """
        G = self.Gs
        if self.qubo != []:
            self.qi()
        J = self.reJ()
        N = len(J)
        q = np.random.choice([-1, 1], self.tro * N).reshape(self.tro, N)
        self.E.append(Ei_sqa(q, J, self.Tf, self.tro, G) + self.ep)
        while G > self.Gf:
            for _ in range(self.ite):
                x = np.random.randint(N)
                y = np.random.randint(self.tro)
                dE = 0

                for j in range(N):
                    if j == x:
                        dE += -2 * q[y][x] * J[x][x] / self.tro
                    else:
                        dE += -2 * q[y][j] * q[y][x] * J[j][x] / self.tro

                dE += q[y][x] * (q[(self.tro + y - 1) % self.tro][x] + q[
                    (y + 1) % self.tro][x]) * np.log(
                        1 / np.tanh(G / self.Tf / self.tro)) * self.Tf

                if dE < 0 or np.exp(-dE / self.Tf) > np.random.rand():
                    q[y][x] *= -1
            self.E.append(Ei_sqa(q, J, self.Tf, self.tro, G) + self.ep)
            G *= self.R
        #qq = [int((i+1)/2) for i in q]
        return q

    def qaoa(self, shots=1, step=2, verbose=False):
        from blueqat import vqe
        return vqe.Vqe(vqe.QaoaAnsatz(pauli(self.qubo), step)).run()

    def dw(self):
        try:
            from dwave.cloud import Client
        except ImportError:
            raise ImportError(
                "dw() requires dwave-cloud-client. Please install before call this function."
            )
        solver = Client.from_config(endpoint=self.dwaveendpoint,
                                    token=self.dwavetoken,
                                    solver=self.dwavesolver).get_solver()

        if self.qubo != []:
            self.qi()

        # for hi
        harr = np.diag(self.J)
        larr = []
        for i in solver.nodes:
            if i < len(harr):
                larr.append(harr[i])
        linear = {index: larr[index] for index in range(len(larr))}

        # for jij
        qarr = []
        qarrv = []
        for i in solver.undirected_edges:
            if i[0] < len(harr) and i[1] < len(harr):
                qarr.append(i)
                qarrv.append(self.J[i[0]][i[1]])

        quad = {key: j for key, j in zip(qarr, qarrv)}
        computation = solver.sample_ising(linear, quad, num_reads=1)

        return list(
            map(lambda s: int((s + 1) / 2),
                computation.samples[0][:len(harr)]))

    def setRBM(self, rbm_arr):
        self.qubo = dbms(rbm_arr)
        self.RBMvisible = rbm_arr[0]
        self.RBMhidden = rbm_arr[1]

    def rbm_model_qubo(self, shots=100, targetT=0.02):
        result = self.run(shots=shots, sampler="fast", targetT=targetT)
        bias = np.sum(result, axis=0) / shots
        bias_qubo = np.diag(bias)
        weight_qubo = zeros(len(bias))
        for i in range(self.RBMvisible):
            for j in range(self.RBMvisible, self.RBMvisible + self.RBMhidden):
                for k in range(len(result)):
                    weight_qubo[i][j] += result[k][i] * result[k][j]
        return bias_qubo + weight_qubo / shots

    def fit(self,
            vdata,
            shots=100,
            targetT=0.02,
            alpha=0.9,
            epsilon=0.1,
            epoch=100,
            verbose=True):
        for i in range(epoch):
            x = np.random.randint(0, len(vdata))
            vdata_in = [int((i - 1) * (-1)) for i in vdata[x]]
            data_qubo = rbm_data_qubo(vdata_in, self.qubo)
            model_qubo = self.rbm_model_qubo(shots=shots, targetT=targetT)
            self.qubo = np.asarray(self.qubo) * alpha + epsilon * (
                np.asarray(data_qubo) - np.asarray(model_qubo))
            if verbose:
                print("epoch", i, ":", self.qubo)




"""Defines JSON serializer and deserializer."""
import typing
from blueqat import Circuit

from ..gateset import create
from .flatten import flatten

SCHEMA_NAME = 'blueqat-circuit'
SCHEMA_VERSION = "1"

if typing.TYPE_CHECKING:
    from blueqat.gate import Operation
    try:
        from typing import Any, Dict, List, TypedDict
    except ImportError:
        CircuitJsonDict = Dict[str, Any]
        OpJsonDict = Dict[str, Any]
    else:
        class SchemaJsonDict(TypedDict):
            """Schema header for detect data type"""
            name: str
            version: str

        class OpJsonDict(TypedDict):
            """Data type of Operation"""
            name: str
            params: List[float]
            targets: List[int]

        class CircuitJsonDict(TypedDict):
            """Data type of Circuit"""
            schema: SchemaJsonDict
            n_qubits: int
            ops: List[OpJsonDict]


def serialize(c: Circuit) -> 'CircuitJsonDict':
    """Serialize Circuit into JSON type dict.

    In this implementation, serialized circuit is flattened.
    However, it's not specifications of JSON schema.
    """
    def serialize_op(op: 'Operation') -> 'OpJsonDict':
        targets = op.targets
        if isinstance(targets, slice):
            raise TypeError('Not flatten circuit.')
        if isinstance(targets, int):
            targets = [targets]
        if isinstance(targets, tuple):
            targets = list(targets)
        return {
            'name': str(op.lowername),
            'params': [float(p) for p in op.params],
            'targets': targets
        }

    c = flatten(c)
    return {
        'schema': {
            'name': SCHEMA_NAME,
            'version': SCHEMA_VERSION
        },
        'n_qubits': c.n_qubits,
        'ops': [serialize_op(op) for op in c.ops]
    }


def deserialize(data: 'CircuitJsonDict') -> Circuit:
    """Deserialize JSON type dict into Circuit"""
    def make_op(opdata: 'OpJsonDict') -> 'Operation':
        return create(opdata['name'],
                      tuple(opdata['targets']),
                      tuple(float(p) for p in opdata['params']))
    schema = data.get('schema', {})
    if schema.get('name', '') != SCHEMA_NAME:
        raise ValueError('Invalid schema')
    if schema.get('version', '') != SCHEMA_VERSION:
        raise ValueError('Unknown schema version')
    n_qubits = data['n_qubits']
    ops = data['ops']
    return Circuit(n_qubits, [
        make_op(opdata) for opdata in ops
    ])
